@@ -1,3 +1,4 @@
+/* vim:set noet ts=8 sw=8 : */
 #include "greenlet.h"
 #include "structmember.h"
 
@@ -77,9 +78,29 @@ The running greenlet's stack_start is undefined but not NULL.
      using the dictionary key 'ts_curkey'.
 */
 
+/* Python 2.3 support */
+#ifndef Py_VISIT
+#define Py_VISIT(o) if (o) {if ((err = visit((PyObject *)(o), arg))) return err;}
+#endif /* !Py_VISIT */
+
+#ifndef Py_CLEAR
+#define Py_CLEAR(op) \
+	do { \
+		if (op) { \
+			PyObject *tmp = (PyObject *)(op); \
+			(op) = NULL; \
+			Py_DECREF(tmp); \
+		} \
+	} while (0)
+#endif /* !Py_CLEAR */
+
+/* The current greenlet in this thread state (holds a reference) */
 static PyGreenlet* ts_current;
+/* Holds a reference to the switching-from stack during the slp switch */
 static PyGreenlet* ts_origin;
+/* Holds a reference to the switching-to stack during the slp switch */
 static PyGreenlet* ts_target;
+/* NULL if error, otherwise args tuple to pass around during slp switch */
 static PyObject* ts_passaround;
 
 /***********************************************************/
@@ -93,6 +114,24 @@ static PyObject* ts_delkey;
 static PyObject* PyExc_GreenletError;
 static PyObject* PyExc_GreenletExit;
 
+#undef GREENLET_USE_GC
+
+#ifdef GREENLET_USE_GC
+#define GREENLET_GC_FLAGS Py_TPFLAGS_HAVE_GC
+#define GREENLET_tp_alloc PyType_GenericAlloc
+#define GREENLET_tp_free PyObject_GC_Del
+#define GREENLET_tp_traverse green_traverse
+#define GREENLET_tp_clear green_clear
+#define GREENLET_tp_is_gc green_is_gc
+#else /* GREENLET_USE_GC */
+#define GREENLET_GC_FLAGS 0
+#define GREENLET_tp_alloc 0
+#define GREENLET_tp_free 0
+#define GREENLET_tp_traverse 0
+#define GREENLET_tp_clear 0
+#define GREENLET_tp_is_gc 0
+#endif /* !GREENLET_USE_GC */
+
 static PyGreenlet* green_create_main(void)
 {
 	PyGreenlet* gmain;
@@ -111,6 +150,9 @@ static PyGreenlet* green_create_main(void)
 	gmain->stack_stop = (char*) -1;
 	gmain->run_info = dict;
 	Py_INCREF(dict);
+#ifdef GREENLET_USE_GC
+	printf("green_create_main %p\n", gmain);
+#endif
 	return gmain;
 }
 
@@ -177,6 +219,7 @@ static int g_save(PyGreenlet* g, char* stop)
 	 */
 	long sz1 = g->stack_saved;
 	long sz2 = stop - g->stack_start;
+	assert(g->stack_start != NULL);
 	if (sz2 > sz1) {
 		char* c = PyMem_Realloc(g->stack_copy, sz2);
 		if (!c) {
@@ -252,7 +295,7 @@ static int slp_save_state(char* stackref)
 #ifndef STACK_MAGIC
 #error "greenlet needs to be ported to this platform,\
  or teached how to detect your compiler properly."
-#endif
+#endif /* !STACK_MAGIC */
 
 
 /* This is a trick to prevent the compiler from inlining or
@@ -330,6 +373,33 @@ static PyObject* g_switch(PyGreenlet* target, PyObject* args)
 	}
 }
 
+static PyObject *g_handle_exit(PyObject *result)
+{
+	if (result == NULL &&
+	    PyErr_ExceptionMatches(PyExc_GreenletExit)) {
+		/* catch and ignore GreenletExit */
+		PyObject *exc, *val, *tb;
+		PyErr_Fetch(&exc, &val, &tb);
+		if (val == NULL) {
+			Py_INCREF(Py_None);
+			val = Py_None;
+		}
+		result = val;
+		Py_DECREF(exc);
+		Py_XDECREF(tb);
+	}
+	if (result != NULL) {
+		/* package the result into a 1-tuple */
+		PyObject* r = result;
+		result = PyTuple_New(1);
+		if (result)
+			PyTuple_SET_ITEM(result, 0, r);
+		else
+			Py_DECREF(r);
+	}
+	return result;
+}
+
 static void g_initialstub(void* mark)
 {
 	int err;
@@ -351,7 +421,10 @@ static void g_initialstub(void* mark)
 	/* start the greenlet */
 	ts_target->stack_start = NULL;
 	ts_target->stack_stop = (char*) mark;
-	ts_target->stack_prev = ts_current;
+	if (ts_current->stack_start == NULL)    /* ts_current is dying */
+		ts_target->stack_prev = ts_current->stack_prev;
+	else
+		ts_target->stack_prev = ts_current;
 	ts_target->top_frame = NULL;
 	ts_target->recursion_depth = PyThreadState_GET()->recursion_depth;
 	err = _PyGreen_switchstack();
@@ -363,7 +436,8 @@ static void g_initialstub(void* mark)
 		/* in the new greenlet */
 		PyObject* args;
 		PyObject* result;
-		ts_current->stack_start = (char*) 1;  /* running */
+		PyGreenlet* ts_self = ts_current;
+		ts_self->stack_start = (char*) 1;  /* running */
 
 		args = ts_passaround;
 		if (args == NULL)    /* pending exception */
@@ -374,33 +448,13 @@ static void g_initialstub(void* mark)
 			Py_DECREF(args);
 		}
 		Py_DECREF(run);
-		if (result == NULL &&
-		    PyErr_ExceptionMatches(PyExc_GreenletExit)) {
-			/* catch and ignore GreenletExit */
-			PyObject *exc, *val, *tb;
-			PyErr_Fetch(&exc, &val, &tb);
-			if (val == NULL) {
-				Py_INCREF(Py_None);
-				val = Py_None;
-			}
-			result = val;
-			Py_DECREF(exc);
-			Py_XDECREF(tb);
-		}
-		if (result != NULL) {
-			/* package the result into a 1-tuple */
-			PyObject* r = result;
-			result = PyTuple_New(1);
-			if (result)
-				PyTuple_SET_ITEM(result, 0, r);
-			else
-				Py_DECREF(r);
-		}
+		result = g_handle_exit(result);
 		/* jump back to parent */
-		ts_current->stack_start = NULL;  /* dead */
-		g_switch(ts_current->parent, result);
+		ts_self->stack_start = NULL;  /* dead */
+		g_switch(ts_self->parent, result);
 		/* must not return from here! */
-		Py_FatalError("XXX memory exhausted at a very bad moment");
+		PyErr_WriteUnraisable((PyObject*) ts_self);
+		Py_FatalError("greenlets cannot continue");
 	}
 	/* back in the parent */
 }
@@ -420,6 +474,9 @@ static PyObject* green_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 		Py_INCREF(ts_current);
 		((PyGreenlet*) o)->parent = ts_current;
 	}
+#ifdef GREENLET_USE_GC
+	printf("green_new %p\n", o);
+#endif
 	return o;
 }
 
@@ -448,11 +505,16 @@ static int kill_greenlet(PyGreenlet* self)
 {
 	/* Cannot raise an exception to kill the greenlet if
 	   it is not running in the same thread! */
+#ifdef GREENLET_USE_GC
+	printf("kill_greenlet %p\n", self);
+#endif
 	if (self->run_info == PyThreadState_GET()->dict) {
 		/* The dying greenlet cannot be a parent of ts_current
 		   because the 'parent' field chain would hold a
 		   reference */
 		PyObject* result;
+		if (!STATE_OK)
+			return -1;
 		Py_INCREF(ts_current);
 		self->parent = ts_current;
 		/* Send the greenlet a GreenletExit exception. */
@@ -482,10 +544,46 @@ static int kill_greenlet(PyGreenlet* self)
 	}
 }
 
+#ifdef GREENLET_USE_GC
+static int
+green_traverse(PyGreenlet *so, visitproc visit, void *arg)
+{
+	printf("green_traverse %p\n", so);
+	Py_VISIT((PyObject*)so->run_info);
+	Py_VISIT((PyObject*)so->parent);
+	Py_VISIT((PyObject*)so->top_frame);
+	Py_VISIT((PyObject*)so->stack_prev);
+	Py_VISIT((PyObject*)so->parent);
+	return 0;
+}
+
+static int green_is_gc(PyGreenlet* self)
+{
+	int rval;
+	rval = (self->stack_stop == ((char *)-1)) ? 0 : 1;
+	printf("green_is_gc %p = %d\n", self, rval);
+	return rval;
+}
+
+static int green_clear(PyGreenlet* self)
+{
+	printf("green_clear %p run_info %p\n", self, self->run_info);
+	if (PyGreen_ACTIVE(self))
+		return kill_greenlet(self);
+	return 0;
+}
+#endif
+
 static void green_dealloc(PyGreenlet* self)
 {
 	PyObject *error_type, *error_value, *error_traceback;
 
+#ifdef GREENLET_USE_GC
+	printf("green_dealloc %p\n", self);
+	PyObject_GC_UnTrack((PyObject *)self);
+	Py_TRASHCAN_SAFE_BEGIN(self);
+#endif /* GREENLET_USE_GC */
+
 	Py_XDECREF(self->parent);
 	self->parent = NULL;
 	if (PyGreen_ACTIVE(self)) {
@@ -522,18 +620,26 @@ static void green_dealloc(PyGreenlet* self)
 			/* Resurrected! */
 			int refcnt = self->ob_refcnt;
 			_Py_NewReference((PyObject*) self);
+#ifdef GREENLET_USE_GC
+			PyObject_GC_Track((PyObject *)self);
+#endif
 			self->ob_refcnt = refcnt;
 #ifdef COUNT_ALLOCS
 			--self->ob_type->tp_frees;
 			--self->ob_type->tp_allocs;
-#endif
-			return;
+#endif /* COUNT_ALLOCS */
+			goto green_dealloc_end;
 		}
 	}
 	if (self->weakreflist != NULL)
 		PyObject_ClearWeakRefs((PyObject *) self);
 	Py_XDECREF(self->run_info);
 	self->ob_type->tp_free((PyObject*) self);
+green_dealloc_end:
+#ifdef GREENLET_USE_GC
+	Py_TRASHCAN_SAFE_END(self);
+#endif /* GREENLET_USE_GC */
+	return;
 }
 
 static PyObject* single_result(PyObject* results)
@@ -549,12 +655,118 @@ static PyObject* single_result(PyObject* results)
 		return results;
 }
 
+static PyObject *
+throw_greenlet(PyGreenlet* self, PyObject* typ, PyObject* val, PyObject* tb)
+{
+	/* Note: _consumes_ a reference to typ, val, tb */
+	PyObject *result = NULL;
+	PyErr_Restore(typ, val, tb);
+	if (PyGreen_STARTED(self) && !PyGreen_ACTIVE(self)) {
+		/* dead greenlet: turn GreenletExit into a regular return */
+		result = g_handle_exit(result);
+	}
+	return single_result(g_switch(self, result));
+}
+PyDoc_STRVAR(green_switch_doc,
+"switch(*args)\n\
+\n\
+Switch execution to this greenlet.\n\
+\n\
+If this greenlet has never been run, then this greenlet\n\
+will be switched to using the body of self.run(*args).\n\
+\n\
+If the greenlet is active (has been run, but was switch()'ed\n\
+out before leaving its run function), then this greenlet will\n\
+be resumed and the return value to its switch call will be\n\
+None if no arguments are given, the given argument if one\n\
+argument is given, or the args tuple if multiple arguments\n\
+are given\n\
+\n\
+If the greenlet is dead, or is the current greenlet then this\n\
+function will simply return the args using the same rules as\n\
+above.");
+
 static PyObject* green_switch(PyGreenlet* self, PyObject* args)
 {
 	Py_INCREF(args);
 	return single_result(g_switch(self, args));
 }
 
+#ifndef PyExceptionClass_Check      /* Python < 2.5 */
+# define PyExceptionClass_Check     PyClass_Check
+#endif
+#ifndef PyExceptionInstance_Check   /* Python < 2.5 */
+# define PyExceptionInstance_Check  PyInstance_Check
+#endif
+#ifndef PyExceptionInstance_Class   /* Python < 2.5 */
+# define PyExceptionInstance_Class(x)				\
+		((PyObject*)((PyInstanceObject*)(x))->in_class)
+#endif
+
+PyDoc_STRVAR(throw_doc,
+"throw(typ[,val[,tb]]) -> raise exception in greenlet, return value passed "
+"when switching back");
+
+static PyObject* green_throw(PyGreenlet* self, PyObject* args)
+{
+	PyObject *typ = PyExc_GreenletExit;
+	PyObject *val = NULL;
+	PyObject *tb = NULL;
+	
+	if (!PyArg_ParseTuple(args, "|OOO:throw", &typ, &val, &tb))
+		return NULL;
+
+	/* First, check the traceback argument, replacing None with
+	   NULL. */
+	if (tb == Py_None)
+		tb = NULL;
+	else if (tb != NULL && !PyTraceBack_Check(tb)) {
+		PyErr_SetString(PyExc_TypeError,
+			"throw() third argument must be a traceback object");
+		return NULL;
+	}
+
+	Py_INCREF(typ);
+	Py_XINCREF(val);
+	Py_XINCREF(tb);
+
+	if (PyExceptionClass_Check(typ)) {
+		PyErr_NormalizeException(&typ, &val, &tb);
+	}
+
+	else if (PyExceptionInstance_Check(typ)) {
+		/* Raising an instance.  The value should be a dummy. */
+		if (val && val != Py_None) {
+			PyErr_SetString(PyExc_TypeError,
+			  "instance exception may not have a separate value");
+			goto failed_throw;
+		}
+		else {
+			/* Normalize to raise <class>, <instance> */
+			Py_XDECREF(val);
+			val = typ;
+			typ = PyExceptionInstance_Class(typ);
+			Py_INCREF(typ);
+		}
+	}
+	else {
+		/* Not something you can raise.  throw() fails. */
+		PyErr_Format(PyExc_TypeError,
+			     "exceptions must be classes, or instances, not %s",
+			     typ->ob_type->tp_name);
+		goto failed_throw;
+	}
+
+	return throw_greenlet(self, typ, val, tb);
+
+failed_throw:
+	/* Didn't use our arguments, so restore their original refcounts */
+	Py_DECREF(typ);
+	Py_XDECREF(val);
+	Py_XDECREF(tb);
+	return NULL;
+}
+
 static int green_nonzero(PyGreenlet* self)
 {
 	return PyGreen_ACTIVE(self);
@@ -665,12 +877,16 @@ PyObject* PyGreen_Current(void)
 
 PyObject* PyGreen_Switch(PyObject* g, PyObject* value)
 {
+	PyGreenlet *self;
 	if (!PyGreen_Check(g)) {
 		PyErr_BadInternalCall();
 		return NULL;
 	}
+	self = (PyGreenlet*) g;
 	Py_XINCREF(value);
-	return single_result(g_switch((PyGreenlet*) g, value));
+	if (PyGreen_STARTED(self) && !PyGreen_ACTIVE(self))
+		value = g_handle_exit(value);
+	return single_result(g_switch(self, value));
 }
 
 int PyGreen_SetParent(PyObject* g, PyObject* nparent)
@@ -686,8 +902,9 @@ int PyGreen_SetParent(PyObject* g, PyObject* nparent)
 
 
 static PyMethodDef green_methods[] = {
-    {"switch", (PyCFunction)green_switch, METH_VARARGS, /*XXXswitch_doc*/ NULL},
-    {NULL,     NULL}		/* sentinel */
+	{"switch", (PyCFunction)green_switch, METH_VARARGS, green_switch_doc},
+	{"throw",  (PyCFunction)green_throw,  METH_VARARGS, throw_doc},
+	{NULL,     NULL}		/* sentinel */
 };
 
 static PyGetSetDef green_getsets[] = {
@@ -718,10 +935,11 @@ static PyNumberMethods green_as_number = {
 
 PyTypeObject PyGreen_Type = {
 	PyObject_HEAD_INIT(NULL)
-	0,
-	"greenlet.greenlet",
-	sizeof(PyGreenlet),
-	0,
+	0,					/* ob_size */
+	"greenlet.greenlet",			/* tp_name */
+	sizeof(PyGreenlet),			/* tp_basicsize */
+	0,					/* tp_itemsize */
+	/* methods */
 	(destructor)green_dealloc,		/* tp_dealloc */
 	0,					/* tp_print */
 	0,					/* tp_getattr */
@@ -737,10 +955,13 @@ PyTypeObject PyGreen_Type = {
 	0,					/* tp_getattro */
 	0,					/* tp_setattro */
 	0,					/* tp_as_buffer*/
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,  /* tp_flags */
-	0,					/* tp_doc */
- 	0,					/* tp_traverse */
-	0,					/* tp_clear */
+	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | GREENLET_GC_FLAGS,	/* tp_flags */
+	"greenlet(run=None, parent=None)\n\
+Create a new greenlet object (without running it).  \"run\" is the\n\
+callable to invoke, and \"parent\" is the parent greenlet, which\n\
+defaults to the current greenlet.",		/* tp_doc */
+	(traverseproc)GREENLET_tp_traverse,	/* tp_traverse */
+	(inquiry)GREENLET_tp_clear,		/* tp_clear */
 	0,					/* tp_richcompare */
 	offsetof(PyGreenlet, weakreflist),	/* tp_weaklistoffset */
 	0,					/* tp_iter */
@@ -754,8 +975,10 @@ PyTypeObject PyGreen_Type = {
 	0,					/* tp_descr_set */
 	0,					/* tp_dictoffset */
 	(initproc)green_init,			/* tp_init */
-	0,					/* tp_alloc */
+	GREENLET_tp_alloc,			/* tp_alloc */
 	green_new,				/* tp_new */
+	GREENLET_tp_free,			/* tp_free */        
+	(inquiry)GREENLET_tp_is_gc,		/* tp_is_gc */
 };
 /* XXX need GC support */
 
@@ -769,8 +992,11 @@ static PyObject* mod_getcurrent(PyObject* self)
 }
 
 static PyMethodDef GreenMethods[] = {
-	{"getcurrent", (PyCFunction)mod_getcurrent, METH_NOARGS, /*XXX*/ NULL},
-	{NULL,     NULL}        /* Sentinel */
+	{"getcurrent", (PyCFunction)mod_getcurrent, METH_NOARGS,
+	"greenlet.getcurrent()\n\
+Returns the current greenlet (i.e. the one which called this\n\
+function)."},
+	{NULL,     NULL}	/* Sentinel */
 };
 
 static char* copy_on_greentype[] = {
@@ -780,11 +1006,21 @@ static char* copy_on_greentype[] = {
 void initgreenlet(void)
 {
 	PyObject* m;
+	PyObject* greenletexit_doc;
+	PyObject* greenletexit_dict;
+	PyObject* greenleterror_doc;
+	PyObject* greenleterror_dict;
+	int error;
 	char** p;
 	_PyGreen_switchstack = g_switchstack;
 	_PyGreen_slp_switch = slp_switch;
 	_PyGreen_initialstub = g_initialstub;
 	m = Py_InitModule("greenlet", GreenMethods);
+	if (m == NULL)
+		return;
+	
+	if (PyModule_AddStringConstant(m, "__version__", "0.2") < 0)
+		return;
 
 	ts_curkey = PyString_InternFromString("__greenlet_ts_curkey");
 	ts_delkey = PyString_InternFromString("__greenlet_ts_delkey");
@@ -792,11 +1028,49 @@ void initgreenlet(void)
 		return;
 	if (PyType_Ready(&PyGreen_Type) < 0)
 		return;
-	PyExc_GreenletError = PyErr_NewException("greenlet.error", NULL, NULL);
+
+        greenleterror_dict = PyDict_New();
+	if (greenleterror_dict == NULL)
+		return;
+	greenleterror_doc = PyString_FromString("internal greenlet error");
+	if (greenleterror_doc == NULL) {
+                Py_DECREF(greenleterror_dict);
+		return;
+        }
+
+        error = PyDict_SetItemString(greenleterror_dict, "__doc__", greenleterror_doc);
+	Py_DECREF(greenleterror_doc);
+	if (error == -1) {
+                Py_DECREF(greenleterror_dict);
+		return;
+        }
+
+	PyExc_GreenletError = PyErr_NewException("py.magic.greenlet.error", NULL, greenleterror_dict);
+        Py_DECREF(greenleterror_dict);
 	if (PyExc_GreenletError == NULL)
 		return;
-	PyExc_GreenletExit = PyErr_NewException("greenlet.GreenletExit",
-						NULL, NULL);
+
+        greenletexit_dict = PyDict_New();
+	if (greenletexit_dict == NULL)
+		return;
+	greenletexit_doc = PyString_FromString("greenlet.GreenletExit\n\
+This special exception does not propagate to the parent greenlet; it\n\
+can be used to kill a single greenlet.\n");
+	if (greenletexit_doc == NULL) {
+                Py_DECREF(greenletexit_dict);
+		return;
+        }
+
+        error = PyDict_SetItemString(greenletexit_dict, "__doc__", greenletexit_doc);
+	Py_DECREF(greenletexit_doc);
+	if (error == -1) {
+                Py_DECREF(greenletexit_dict);
+		return;
+        }
+
+	PyExc_GreenletExit = PyErr_NewException("py.magic.greenlet.GreenletExit",
+						NULL, greenletexit_dict);
+        Py_DECREF(greenletexit_dict);
 	if (PyExc_GreenletExit == NULL)
 		return;
 
@@ -810,8 +1084,13 @@ void initgreenlet(void)
 	PyModule_AddObject(m, "error", PyExc_GreenletError);
 	Py_INCREF(PyExc_GreenletExit);
 	PyModule_AddObject(m, "GreenletExit", PyExc_GreenletExit);
+#ifdef GREENLET_USE_GC
+	PyModule_AddObject(m, "GREENLET_USE_GC", PyBool_FromLong(1));
+#else
+	PyModule_AddObject(m, "GREENLET_USE_GC", PyBool_FromLong(0));
+#endif
 
-        /* also publish module-level data as attributes of the greentype. */
+	/* also publish module-level data as attributes of the greentype. */
 	for (p=copy_on_greentype; *p; p++) {
 		PyObject* o = PyObject_GetAttrString(m, *p);
 		if (!o) continue;
